<?php


define ('POLL_VOTED_NOTIFICATION', 'Благодаря, че гласува!');

define ('POLL_ERROR_EMPTY_FIELD', 'Оставил/а си празно полето: ');


class PollField {
    public $type;
    public $title;
    public $choices;
    public $value;
    public $required;
    function __construct ($new_type)
    {
        $this->type = $new_type;
    }
}


class PollSummary {
    public $votes_count;
    public $votes = [];
}


function make_poll_radio_field ($block_items)
{
  if (count ($block_items) < 1) return FALSE;
  if (count ($block_items) < 3) return FALSE;

  $field = new PollField ('radio');
  $field->title = $block_items[0];
  $field->choices = [];

  for ($i = 1; $i < count ($block_items); ++$i)
    {
      $label = $block_items[$i];
      array_push ($field->choices, $label);
    }

  return $field;
}


function make_poll_text_field ($block_items)
{
    if (count ($block_items) < 1) return FALSE;
    if (count ($block_items) > 1) return FALSE;

    $field = new PollField ('text');
    $field->title = $block_items[0];
    return $field;
}


function get_poll ($filepath)
{
    $file = fopen ($filepath, 'r');
    if ($file === FALSE) return FALSE;

    $fields = [];
    $has_required_fields = FALSE;

    $line = NULL;
    while ($line !== FALSE)
    {
        $block = get_block ($file, $line);
        if ($block === NULL) break;
        if ($block === FALSE) return FALSE;

        $required = FALSE;

        switch ($block->title)
        {
        case 'radio:': $required = TRUE;
        case 'radio-optional:':
            $field = make_poll_radio_field ($block->items);
            break;

        case 'text:': $required = TRUE;
        case 'text-optional:':
            $field = make_poll_text_field ($block->items);
            break;

        default:
            // poll_error ("Unknown poll block: $block->type");
            $field = FALSE;
        }

        if ($field === FALSE) break;

        if ($required) $has_required_fields = TRUE;
        $field->required = $required;
        array_push ($fields, $field);
    }

    fclose ($file);

    if (!$has_required_fields)
    {
        // poll_error ("Must have at least 1 non-optional field.");
        return FALSE;
    }

    return $fields;
}


function get_poll_form ($poll_name, $fields)
{
  $html  = "<article id=\"poll\">\n";
  $html .= "<h2>Кажи си мнението!</h2>\n";
  $html .= "<form method=\"post\" action=\".\">\n";
  $html .= "<input type=\"hidden\" name=\"poll_id\" value=\"$poll_name\">\n";
  $html .= "<ul>\n";

  for ($field_index = 0; $field_index < count ($fields); ++$field_index)
    {
      $field = $fields[$field_index];
      $name = "poll_field_$field_index";
      $required = $field->required ? ' required' : '';

      switch ($field->type)
        {
        case "radio":
          $html .= "<li><p>$field->title</p>\n";
          $html .= "  <ul>\n";
          for ($choice_index = 0; $choice_index < count ($field->choices); ++$choice_index)
            {
              $label = $field->choices[$choice_index];
              $checked = $field->value === $choice_index ? ' checked' : '';
              $html .= "    <li><label><input type=\"radio\" name=\"$name\" value=\"$choice_index\"$checked$required/> $label</label></li>\n";
            }
          $html .= "  </ul>\n";
          $html .= "</li>\n";
          break;
        case "text":
          $value = $field->value ? " value=\"$field->value\"" : '';
          $placeholder = $field->required ? '' : " placeholder=\"(по желание)\"";
          $html .= "<li><label>$field->title<input type=\"text\" name=\"$name\"$value$placeholder$required/></label></li>\n";
          break;
        }
    }

  $html .= "</ul>\n";
  $html .= "<p><button type=\"submit\">Давай!</button></p>\n";
  $html .= "</form>\n";
  $html .= "</article>\n";

  return $html;
}


function get_poll_submit_errors ($fields, $data)
{
  $errors = [];

  for ($field_index = 0; $field_index < count ($fields); ++$field_index)
    {
      $field = $fields[$field_index];
      $name = "poll_field_$field_index";
      $required = $field->required;

      if (isset ($data[$name]))
        {
          $value = trim ($data[$name]);

          if ($value === '')
            {
              if ($required) array_push ($errors, POLL_ERROR_EMPTY_FIELD . $field->title);
              continue;
            }

          switch ($field->type)
            {
              case 'radio':
                {
                  if (ctype_digit ($value) && $value < count ($field->choices))
                    {
                      $field->value = (int) $value;
                    }
                  else
                    {
                      array_push ($errors, "Invalid radio value '$value' for poll field: " . $field->title);
                    }
                } break;
              case 'text':
                {
                  $field->value = $value;
                } break;
            }
        }
      else if ($required)
        {
          array_push ($errors, 'A requied poll field is missing: ' . $field->title);
        }
    }

  return $errors;
}


function get_poll_summary_fd ($file, $fields)
{
    $summary = new PollSummary ();

    $summary->votes_count = fgets ($file);
    if ($summary->votes_count === FALSE) return FALSE;
    $summary->votes_count = (int) $summary->votes_count;

    foreach ($fields as $field)
    {
        if ($field->type === 'radio')
        {
            $votes = fgets ($file);
            if ($votes === FALSE) return FALSE;
            $votes = explode (' ', $votes);
            if (count ($votes) !== count ($field->choices)) return FALSE;

            for ($i = 0; $i < count ($votes); ++$i)
            {
                $votes[$i] = (int) $votes[$i];
            }

            array_push ($summary->votes, $votes);
        }
    }

    return $summary;
}


function make_new_poll_summary ($fields)
{
  $summary = new PollSummary ();
  $summary->votes_count = 0;

  foreach ($fields as $field)
    {
      if ($field->type === 'radio')
        {
          $votes = [];

          foreach ($field->choices as $choice)
            {
              array_push ($votes, 0);
            }

          array_push ($summary->votes, $votes);
        }
    }

  return $summary;
}


function get_poll_summary ($filepath, $fields)
{
  if (file_exists ($filepath))
    {
      $file = fopen ($filepath, 'r');
      if ($file === FALSE) return FALSE;
      $locked = flock ($file, LOCK_SH);
      if ($locked === FALSE) return FALSE;

      $summary = get_poll_summary_fd ($file, $fields);
      if ($summary === FALSE) return FALSE;

      fclose ($file);
    }
  else
    {
      $summary = make_new_poll_summary ($fields);
    }

  return $summary;
}


function show_poll_summary ($summary, $fields)
{
  $html  = "<article id=\"poll\" class=\"results\">\n";
  $html .= "  <ul>\n";
  $votes_index = 0;

  foreach ($fields as $field)
    {
      if ($field->type === 'radio')
        {
          $votes = $summary->votes[$votes_index++];

          $html .= "<li>\n";
          $html .= "$field->title\n";
          $html .= "<ul>\n";
          // $html .= "<li><meter value=\"$value\" max=\"$max\"></meter></li>\n";
          $total_votes = array_sum ($votes);
          for ($i = 0; $i < count($field->choices); ++$i)
            {
              $percentage = $total_votes ? (int) ($votes[$i] * 100 / $total_votes) : 0;
              $choice = $field->choices[$i];
              $html .= "<li>$choice: $percentage%</li>\n";
            }
          $html .= "</ul>\n";
          $html .= "</li>\n";
        }
    }
  $html .= "  </ul>\n";
  $html .= "</article>\n";

  return $html;
}


function poll_check_if_voted_fd ($file, $voter_id)
{
  while (TRUE)
    {
      $line = fgets ($file);
      if ($line === FALSE) break;
      if (trim ($line) === $voter_id) return TRUE;
    }

  return FALSE;
}


function poll_check_if_voted ($filepath, $voter_id)
{
  if (file_exists ($filepath))
    {
      $file = fopen ($filepath, 'r+');
      if ($file === FALSE) return NULL;
      $locked = flock ($file, LOCK_SH);
      if ($locked === FALSE) return NULL;

      $has_voted = poll_check_if_voted_fd ($file, $voter_id);

      fclose ($file);
      return $has_voted;
    }
  else
    {
      return FALSE;
    }
}


function save_poll_summary ($filepath, $fields)
{
  $file = fopen ($filepath, 'c+');
  if ($file === FALSE) return FALSE;
  $locked = flock ($file, LOCK_EX);
  if ($locked === FALSE) return FALSE;

  $summary = get_poll_summary_fd ($file, $fields);
  if ($summary === FALSE)
    {
      $summary = make_new_poll_summary ($fields);
    }

  if ($summary === FALSE) return FALSE;

  ++$summary->votes_count;

  $votes_index = 0;
  foreach ($fields as $field)
    {
      if ($field->type === 'radio')
        {
          if ($field->value !== NULL)
            {
              ++$summary->votes[$votes_index][$field->value];
            }
          ++$votes_index;
        }
    }

  ftruncate ($file, 0);
  rewind ($file);
  fwrite ($file, "$summary->votes_count\n");
  foreach ($summary->votes as $votes)
    {
      $line = implode (' ', $votes) . "\n";
      fwrite ($file, $line);
    }

  fclose ($file);

  return $summary;
}


function save_poll_result ($filepath, $fields)
{
    $file = fopen ($filepath, 'a');
    if ($file === FALSE) return FALSE;
    $locked = flock ($file, LOCK_EX);
    if ($locked === FALSE) return FALSE;

    for ($i = 0; $i < count ($fields); ++$i)
    {
        $field = $fields[$i];
        fwrite ($file, "$i:$field->value\n");
    }

    fclose ($file);
}


function save_poll_vote ($results_filepath, $summary_filepath, $voters_filepath, $voter_id, $fields)
{
  $voters_file = fopen ($voters_filepath, 'c+');
  if ($voters_file === FALSE) return FALSE;
  $voters_locked = flock ($voters_file, LOCK_EX);
  if ($voters_locked === FALSE) return FALSE;

  $has_voted = poll_check_if_voted_fd ($voters_file, $voter_id);

  if ($has_voted)
    {
      $summary = NULL;
    }
  else
    {
      $status = save_poll_result ($results_filepath, $fields);
      if ($status === FALSE) return FALSE;
      $status = save_poll_summary ($summary_filepath, $fields);
      if ($status === FALSE) return FALSE;

      fwrite ($voters_file, $voter_id . "\n");

      $summary = $status;
    }

  fclose ($voters_file);

  return $summary;
}


function use_poll ()
{
  if (!isset ($_SERVER['REMOTE_ADDR'])) return FALSE;

  $poll_filepath    = POLLS_DIRPATH . '/'         . ACTIVE_POLL;
  $results_filepath = POLLS_DIRPATH . '/results/' . ACTIVE_POLL;
  $summary_filepath = POLLS_DIRPATH . '/results/' . ACTIVE_POLL . '-summary';
  $voters_filepath  = POLLS_DIRPATH . '/results/' . ACTIVE_POLL . '-voters';

  $poll = get_poll ($poll_filepath);
  if ($poll === FALSE) return FALSE;

  $user_agent = isset ($_SERVER['HTTP_USER_AGENT']) ? $_SERVER['HTTP_USER_AGENT'] : '';
  $voter_id = hash ('sha512', ACTIVE_POLL . $_SERVER['REMOTE_ADDR'] . $user_agent . POLLS_SALT);
  $has_voted = poll_check_if_voted ($voters_filepath, $voter_id);

  if ($has_voted)
    {
      $html = show_poll_summary (get_poll_summary ($summary_filepath, $poll), $poll);
    }
  else if (isset ($_POST['poll_id']) && $_POST['poll_id'] === ACTIVE_POLL)
    {
      $errors = get_poll_submit_errors ($poll, $_POST);
      if ($errors)
        {
          foreach ($errors as $error) push_notification ($error);
          $html = get_poll_form (ACTIVE_POLL, $poll);
        }
      else
        {
          $summary = save_poll_vote ($results_filepath, $summary_filepath, $voters_filepath, $voter_id, $poll);
          if ($summary === FALSE)
            {
              push_notification ('Your vote wasn\'t saved due to technical problems on our side.');
              $html = get_poll_form (ACTIVE_POLL, $poll);
            }
          else
            {
              push_notification (POLL_VOTED_NOTIFICATION);
              $html = show_poll_summary ($summary, $poll);
            }
        }
    }
  else
    {
      $html = get_poll_form (ACTIVE_POLL, $poll);
    }

  return $html;
}

?>
